/***************************************************************************
 *   Copyright (C) 2004 by Stefano Salvador                                *
 *   salva_ste@tin.it                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qdatetime.h>
#include <qdom.h>
#include <qmap.h>

#include <kmimetype.h>
#include <ktempfile.h>
#include <kdebug.h>
#include <kio/job.h>
#include <kfileitem.h>
#include <kio/netaccess.h>
#include <kfilterdev.h>
#include <klocale.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <kinstance.h>

#include "katalog.h"

using namespace KIO;

Katalog::Katalog()
{}

Katalog::~Katalog()
{}

int Katalog::initDocument(const KURL& url)
{
  QString tmpfile;
  QIODevice * dev = 0L;
  bool newFile = FALSE;
  if( !url.isEmpty() && NetAccess::download( url, tmpfile, 0 ) )
  {
    KMimeType::Ptr mmp = KMimeType::findByURL(url);
    dev = KFilterDev::deviceForFile( tmpfile, mmp->name());
    if(!dev->open( IO_ReadOnly ))
      return UNREADABLE_URL;
  }
  else
    newFile = TRUE;
  
  m_document = QDomDocument();
  if(newFile)
  {
    m_document = QDomDocument("KDEKatalog");
    m_rootElement = m_document.createElement("KATALOG");
    m_document.appendChild(m_rootElement);
  }
  else
  {
    int Line;
    int Col;
    QString str;
    if ( !dev || 
         !m_document.setContent( dev, true, &str, &Line, &Col ) ||
         m_document.doctype().name().compare("KDEKatalog"))
    {
      dev->close();
      NetAccess::removeTempFile( tmpfile );
      return NOT_KATALOG;
    }
      
    dev->close();
    NetAccess::removeTempFile( tmpfile );
      
    QDomNode node = m_document.firstChild();
    while ( !node.isNull() &&
            !node.isElement() &&
            node.nodeName() != "KATALOG")
      node = node.nextSibling();
    
    if(node.isNull())
      return UNCOMPLETE;
    
    m_rootElement = node.toElement();
  }
  
  return SUCCESS;
}

bool Katalog::saveDocument(const KURL& file, const char *format)
{
  if(!file.isValid())
    return FALSE;
    
  QString tmpName;
  QIODevice *dev;
  if(!file.isLocalFile())
  {
    KTempFile tmpFile;
    tmpName = tmpFile.name();
    dev = KFilterDev::deviceForFile( tmpName, format );
  }
  else
    dev = KFilterDev::deviceForFile( file.path(), format );
    
  if(!dev->open( IO_WriteOnly ))
    return FALSE;
  
  QTextStream ds( dev );
  ds << m_document.toString(2);
  dev->close();
  delete dev;
  
  if(!file.isLocalFile())
    NetAccess::upload( tmpName, file, 0 );
  
  return TRUE;
}
 
int Katalog::addItems(const KURL& mount, const char *name, const QDateTime *date, bool exploreArchives)
{
  m_exploreArchives = exploreArchives;
  
  if(name == 0 || *name == 0)
    return NAME_EMPTY;
    
  if(!date)
    date = new QDateTime(QDate::currentDate(), QTime::currentTime());
    
  ListJob *job = listRecursive( mount, false );
  QStringList list(name);
  QDomNode node = findNode(list);
  QDomElement el;
  if(node.isNull())
  {
    el = m_document.createElement("CATALOG");
    el.setAttribute("name", name);
    el.setAttribute("type", E_DIR);
    el.setAttribute("mimetype", "inode/directory");
    el.setAttribute("time", date->toTime_t());
    m_rootElement.appendChild(el);
  }
  else
    el = node.toElement();
    
  delete date;
      
  m_map[job->url()] = QStringList(name);
  
  connect( job, SIGNAL( entries( KIO::Job *, const KIO::UDSEntryList & ) ),
           this, SLOT( slotEntries( KIO::Job *, const KIO::UDSEntryList & ) ) );
  connect( job, SIGNAL( result( KIO::Job * ) ),
           this, SLOT( slotResult( KIO::Job * ) ) );
  connect( job, SIGNAL( redirection( KIO::Job *, const KURL& ) ),
           this, SLOT( slotRedirection( KIO::Job *, const KURL& ) ) );
  
  return SUCCESS;
}

UDSEntry Katalog::findEntry(QStringList& path) const
{
  QDomNode node = findNode(path);
  if(node == m_rootElement)
  {
    UDSEntry entry;
    UDSAtom atom;
    atom.m_uds = UDS_FILE_TYPE;
    atom.m_long = S_IFDIR;
    entry.append( atom );
    return entry;
  }
  
  if(node.isNull()){
    UDSEntry entry;
    entry.clear();
    return entry;
  }
  
  return createUDSEntry(node.toElement());
}

UDSEntryList Katalog::getNodeContent(QStringList& path) const
{
  UDSEntryList l;
  QDomNode parentNode = findNode(path);
  QDomNode node = parentNode.firstChild();
  while ( !node.isNull() )
  {
    if(!node.isElement())
      continue;
    l.append( createUDSEntry(node.toElement()) );
    node = node.nextSibling();
  }
  return l;
}

QDomNode Katalog::findNode(QStringList& path) const
{
  // If path is long 0 return the root node
  if(path.size() == 0)
    return m_rootElement;
  
  // Find the catalog
  QStringList::Iterator it = path.begin();
  
  QDomNode node = m_rootElement.firstChild();
  
  while ( !node.isNull())
  { 
    if ( node.nodeName() == "CATALOG" &&
        node.isElement() &&
        node.toElement().attribute( "name" ) == *it)
      break;
    node = node.nextSibling();
  }
    
  // If path is long 1 return the node of the catalog
  if(path.size() == 1)
    return node;
  
  // Find the specific node
  while(++it != path.end())
  {
    node = node.firstChild();
    while ( !node.isNull())
    { 
      if ( node.nodeName() == "ITEM" &&
          node.isElement() &&
          node.toElement().attribute( "name" ) == *it)
        break;
      node = node.nextSibling();
    }
  }
  return node;
}

void Katalog::slotEntries(Job *job, const UDSEntryList &entries)
{
  KURL url = static_cast<ListJob *>(job)->url();
  
  QMapIterator<KURL, QStringList> mit = m_map.find(url);
  if(mit == m_map.end())
    return;
    
  QStringList path = *mit;
  
  UDSEntryListConstIterator it = entries.begin();
  UDSEntryListConstIterator end = entries.end();
  for ( ; it != end; ++it )
  {
    QString name;

    UDSEntry::ConstIterator entit = (*it).begin();
    for( ; entit != (*it).end(); ++entit )
      if ( (*entit).m_uds == UDS_NAME )
      {
        name = (*entit).m_str;
        break;
      }

    if ( name.isEmpty() || name == "." || name == "..")
      continue;
    
    KFileItem item( *it, url, false, true );
    
    QStringList tmpPath = path + QStringList::split("/", name);
    QString nameEl = tmpPath.last();
    tmpPath.remove(tmpPath.last());

    QDomElement childNode = m_document.createElement("ITEM");
    QDateTime date;
    date.setTime_t( item.time( UDS_MODIFICATION_TIME ));
    childNode.setAttribute("name", nameEl);
    if(item.isDir())
      childNode.setAttribute("type", E_DIR);
    else if(item.isFile())
      childNode.setAttribute("type", E_FILE);
    else if(item.isLink())
      childNode.setAttribute("type", E_LINK);
    childNode.setAttribute("mimetype", item.mimetype()); 
    childNode.setAttribute("time", date.toTime_t());
    childNode.setAttribute("size", number( item.size() ));
    
    QDomNode node = findNode(tmpPath);
    node.appendChild(childNode);
    
    bool list = false;
    KURL subUrl;
    if(item.mimetype() == "application/x-tgz" || 
       item.mimetype() == "application/x-tbz" || 
       item.mimetype() == "application/x-tar") {
      subUrl.setProtocol("tar");
      list = true;
    } else if(item.mimetype() == "application/x-zip") {
      subUrl.setProtocol("zip");
      list = true;
    }
    
    if(list && m_exploreArchives){
      subUrl.setPath( item.url().path());
      ListJob *subjob = listRecursive( subUrl, false );
      tmpPath.append( nameEl );
      m_map[subjob->url()] = tmpPath;
      connect( subjob, SIGNAL( entries( KIO::Job *, const KIO::UDSEntryList & ) ),
               this, SLOT( slotEntries( KIO::Job *, const KIO::UDSEntryList & ) ) );
      connect( subjob, SIGNAL( result( KIO::Job * ) ),
               this, SLOT( slotResult( KIO::Job * ) ) );
      connect( subjob, SIGNAL( redirection( KIO::Job *, const KURL& ) ),
               this, SLOT( slotRedirection( KIO::Job *, const KURL& ) ) );
    }
  }
}

void Katalog::slotResult(Job *job)
{ 
  m_map.remove(static_cast<ListJob *>(job)->url());
  if(m_map.isEmpty())
    emit finished();
}

void Katalog::slotRedirection(Job *job, const KURL& newUrl)
{ 
  KURL url = static_cast<ListJob *>(job)->url();
  QMapIterator<KURL, QStringList> mit = m_map.find(url);
  if(mit == m_map.end())
    return;
  
  QStringList path = *mit;
  m_map.remove(url);
  m_map[newUrl] = path;
}

UDSEntry Katalog::createUDSEntry(QDomElement node) const
{
  UDSEntry entry;
  entry.clear();
  int type = node.toElement().attribute( "type" ).toInt();
  if(!node.firstChild().isNull())
    type = E_DIR;
   
  UDSAtom atom;
  atom.m_uds = UDS_NAME;
  atom.m_str = node.toElement().attribute("name");
  entry.append( atom );
  
  atom.m_uds = UDS_FILE_TYPE;
  switch(type)
  {
    case E_FILE:
      atom.m_long = S_IFREG;
      break;
    case E_DIR:
      atom.m_long = S_IFDIR;
      break;
    case E_LINK:
      atom.m_long = S_IFLNK;
      break;
  }
  entry.append( atom );
  
  atom.m_uds = UDS_MODIFICATION_TIME;
  atom.m_long = node.toElement().attribute( "time" ).toInt();
  entry.append( atom );
  
  atom.m_uds = UDS_SIZE;
  atom.m_long = node.toElement().attribute( "size" ).toInt();
  entry.append( atom );
  
  atom.m_uds = UDS_MIME_TYPE;
  if(type == E_DIR)
    atom.m_str = "inode/directory";
  else
    atom.m_str = node.toElement().attribute("mimetype");
  entry.append( atom );
  
  return entry;
}

#include "katalog.moc"
